﻿namespace TabsStudioSorter
{
    class SuperGroup
    {
        public SuperGroup(Options options, PriorityGroup priorityGroup, Marker marker,
            TabsStudioExt.Util.IsPreviewTabProperty isPreviewTabProperty,
        System.Collections.Generic.Dictionary<System.Windows.Controls.TabItem, FixedTabInfo> fixedInfo)
        {
            this.options = options;
            this.priorityGroup = priorityGroup;
            this.marker = marker;
            this.isPreviewTabProperty = isPreviewTabProperty;
            this.fixedInfo = fixedInfo;
            superGroupProperties = new TabsStudioExt.Util.SuperGroupProperties();
        }

        public void UpdateProperties(System.Windows.Controls.Panel panel)
        {
            foreach (System.Windows.Controls.TabItem i in panel.Children)
            {
                string name = superGroupProperties.GetName(i);
                if (name != null && name.StartsWith(superGroupName))
                {
                    superGroupProperties.SetName(i, null);
                    superGroupProperties.SetFirstTab(i, false);
                    superGroupProperties.SetLastTab(i, false);
                }
            }

            foreach (SortingRule rule in options.rules)
            {
                if (rule.superGroup)
                    CreateProperties(panel, rule);
            }
        }

        public void UpdateDeletedProperties(System.Windows.Controls.TabItem deletedTabItem,
            System.Windows.Controls.Panel panel)
        {
            if (superGroupProperties.IsFirstTab(deletedTabItem))
            {
                int nextTabIndex = panel.Children.IndexOf(deletedTabItem) + 1;
                if (nextTabIndex < panel.Children.Count)
                {
                    System.Windows.Controls.TabItem nextTab = TabItemChild(panel, nextTabIndex);
                    if (SameGroup(deletedTabItem, nextTab))
                        superGroupProperties.SetFirstTab(nextTab, true);
                }
            }
            if (superGroupProperties.IsLastTab(deletedTabItem))
            {
                int prevTabIndex = panel.Children.IndexOf(deletedTabItem) - 1;
                if (prevTabIndex >= 0)
                {
                    System.Windows.Controls.TabItem prevTab = TabItemChild(panel, prevTabIndex);
                    if (SameGroup(deletedTabItem, prevTab))
                        superGroupProperties.SetLastTab(prevTab, true);
                }
            }
        }

        private bool SameGroup(System.Windows.Controls.TabItem lhs, System.Windows.Controls.TabItem rhs)
        {
            return superGroupProperties.GetName(lhs) == superGroupProperties.GetName(rhs);
        }

        private void CreateProperties(System.Windows.Controls.Panel panel, SortingRule rule)
        {
            for (int i = 0; i < panel.Children.Count; ++i)
            {
                System.Windows.Controls.TabItem tab = TabItemChild(panel, i);
                if (isPreviewTabProperty.IsPreviewTab(tab) || priorityGroup.IsPriorityTab(tab))
                    continue;

                if (Match(tab, rule) != null)
                {
                    superGroupProperties.SetName(tab, SuperGroupRuleName(rule));
                    superGroupProperties.SetFirstTab(tab,
                        i == 0 || priorityGroup.IsPriorityTab(TabItemChild(panel, i - 1)) || !SameMatch(tab, TabItemChild(panel, i - 1), rule));
                    superGroupProperties.SetLastTab(tab,
                        i == panel.Children.Count - 1 || !SameMatch(tab, TabItemChild(panel, i + 1), rule));
                }
            }
        }

        private object Match(System.Windows.Controls.TabItem tab, SortingRule rule)
        {
            if (!fixedInfo.ContainsKey(tab))
                return null;

            FixedTabInfo info = fixedInfo[tab];
            switch (rule.type)
            {
                case SortingRuleType.MarkerHighlightIndex:
                    uint result = marker.TabHighlightIndex(tab);
                    if (result == 0)
                        return null;
                    else
                        return result;
                case SortingRuleType.Project:
                    return RuleMatch.ExtractM(info.project, rule);
                case SortingRuleType.Extension:
                    return RuleMatch.ExtractM(info.extension, rule);
                case SortingRuleType.TabName:
                    return RuleMatch.ExtractM(RuleMatch.GetTabNameForSorting(tab, false), rule);
                case SortingRuleType.TabNameWithPath:
                    return RuleMatch.ExtractM(RuleMatch.GetTabNameForSorting(tab, true), rule);
                case SortingRuleType.TabNameUntransformed:
                    return RuleMatch.ExtractM(info.untransformedName, rule);
                case SortingRuleType.IsDocument:
                    return info.isDocument;
                case SortingRuleType.Path:
                    return RuleMatch.ExtractM(info.paths, rule);
            }
            return null;
        }

        private bool SameMatch(System.Windows.Controls.TabItem tab, System.Windows.Controls.TabItem otherTab, SortingRule rule)
        {
            object o = Match(tab, rule);
            object other = Match(otherTab, rule);
            if (o is string && other is string)
                return System.String.Compare(o as string, other as string, true) == 0;
            return o.Equals(other);
        }

        private string SuperGroupRuleName(SortingRule rule)
        {
            return superGroupName + " " + rule.name;
        }

        private System.Windows.Controls.TabItem TabItemChild(System.Windows.Controls.Panel panel, int i)
        {
            return (System.Windows.Controls.TabItem)panel.Children[i];
        }

        private Options options;
        private PriorityGroup priorityGroup;
        private Marker marker;
        private TabsStudioExt.Util.IsPreviewTabProperty isPreviewTabProperty;
        private System.Collections.Generic.Dictionary<System.Windows.Controls.TabItem, FixedTabInfo> fixedInfo;
        private TabsStudioExt.Util.SuperGroupProperties superGroupProperties;
        private readonly string superGroupName = "[Sorter]";
    }
}
